<?php
/*
Plugin Name: Payflex Payment Gateway
Description: Use Payflex as a credit card processor for WooCommerce.
Version: 2.3.0
Author: Payflex
*/

/**
 * Check if WooCommerce is activated
 */
 $plugin_path = trailingslashit( WP_PLUGIN_DIR ) . 'woocommerce/woocommerce.php';
if (in_array( $plugin_path, wp_get_active_and_valid_plugins())) {
    
	add_action('plugins_loaded', 'partpay_gateway', 0);

	function partpay_gateway() {

		if (!class_exists('WC_Payment_Gateway'))  return;

		class WC_Gateway_PartPay extends WC_Payment_Gateway {

			/**
			 * @var $_instance WC_Gateway_PartPay The reference to the singleton instance of this class
			 */
			private static $_instance = NULL;

			/**
			 * @var boolean Whether or not logging is enabled
			 */
			public static $log_enabled = false;

			/**
			 * @var WC_Logger Logger instance
			 */
			public static $log = false;


			/**
			 * Main WC_Gateway_PartPay Instance
			 *
			 * Used for WP-Cron jobs when
			 *
			 * @since 1.0
			 * @return WC_Gateway_PartPay Main instance
			 */
			public static function instance() {
				if (is_null(self::$_instance)) {
					self::$_instance = new self();
				}
				return self::$_instance;
			}

			public function __construct() {

				$this->id 					= 'payflex';
				$this->method_title 		= __('Payflex', 'woo_partpay');
				$this->method_description 	= __('Use Payflex as a credit card processor for WooCommerce.', 'woo_partpay');
				$this->icon 				= WP_PLUGIN_URL . "/" . plugin_basename( dirname(__FILE__)) . '/Checkout.png';

				$this->supports 			= array( 'products', 'refunds' );
				
				// Load the form fields.
				$this->init_environment_config();

				// Load the form fields.
				$this->init_form_fields();

				// Load the settings.
				$this->init_settings();

				// Load the frontend scripts.
				$this->init_scripts_js();
				$this->init_scripts_css();
				$settings = get_option('woocommerce_payflex_settings');
				$api_url;

				if( false !== $settings ){
					$api_url  = $this->environments[$this->settings['testmode']]['api_url'];
					$this->orderurl = $api_url . '/order';
					$this->configurationUrl = $api_url . '/configuration';
				} else {
					$api_url = '';
				}
				// Define user set variables
				$this->title = '';
				if (isset($this->settings['title'])) {
					$this->title = $this->settings['title'];
				}
				$this->description = __('Pay with any Visa or Mastercard.','woo_partpay');

				self::$log_enabled	= true;

				// Hooks
				add_action( 'woocommerce_receipt_'.$this->id, array($this, 'receipt_page') );

				add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, array( $this, 'process_admin_options' ) );

				//add_filter( 'woocommerce_thankyou_order_id',array($this,'payment_callback'));

				add_action( 'woocommerce_order_status_refunded',array($this,'create_refund'));

				// Don't enable PartPay if the amount limits are not met
				add_filter('woocommerce_available_payment_gateways',array($this,'check_cart_within_limits'), 99, 1);

				add_action('woocommerce_settings_start', array($this,'update_payment_limits'));
			}


			/**
			 * Initialise Gateway Settings Form Fields
			 *
			 * @since 1.0.0
			 */
			function init_form_fields() {

				$env_values = array();
				foreach( $this->environments as $key => $item ) {
					$env_values[$key] = $item["name"];
				}

				$this->form_fields = array(
					'enabled' => array(
						'title' => __( 'Enable/Disable', 'woo_partpay' ),
						'type' => 'checkbox',
						'label' => __( 'Enable Payflex', 'woo_partpay' ),
						'default' => 'yes'
					),
					'title' => array(
						'title' => __( 'Title', 'woo_partpay' ),
						'type' => 'text',
						'description' => __( 'This controls the payment method title which the user sees during checkout.', 'woo_partpay' ),
						'default' => __( 'Payflex', 'woo_partpay' )
					),
					'testmode' => array(
						'title' => __( 'Test mode', 'woo_partpay' ),
						'label' => __( 'Enable Test mode', 'woo_partpay' ),
						'type' => 'select',
						'options' => $env_values,
						'description' => __( 'Process transactions in Test/Sandbox mode. No transactions will actually take place.', 'woo_partpay' ),
					),
					'client_id' => array(
						'title' => __( 'Client ID', 'woo_partpay' ),
						'type' => 'text',
						'description' => __( 'Payflex Client ID credential', 'woo_partpay' ),
						'default' => __( '', 'woo_partpay' )
					),
					'client_secret' => array(
						'title' => __( 'Client Secret', 'woo_partpay' ),
						'type' => 'text',
						'description' => __( 'Payflex Client Secret credential', 'woo_partpay' ),
						'default' => __( '', 'woo_partpay' )
					),
					'enable_product_widget' => array(
						'title' => __( 'Product Page Widget', 'woo_partpay' ),
						'type' => 'checkbox',
						'label' => __( 'Enable Product Page Widget', 'woo_partpay' ),
						'default' => 'yes',

					),
					'is_using_page_builder' => array(
						'title' => __( 'Product Page Widget using any page builder', 'woo_partpay' ),
						'type' => 'checkbox',
						'label' => __( 'Enable Product Page Widget using page builder', 'woo_partpay' ),
						'default' => 'no',
						'description' => __( '<h3 class="wc-settings-sub-title">Page Builders</h3> If you use a page builder plugin, the above payment info can be placed using a shortcode instead of relying on hooks. Use [payflex_widget] within a product page.', 'woo_partpay' )

					),
					'enable_checkout_widget' => array(
						'title' => __( 'Checkout Page Widget', 'woo_partpay' ),
						'type' => 'checkbox',
						'label' => __( 'Enable Checkout Page Widget', 'woo_partpay' ),
						'default' => 'yes'
					),
					'merchant_widget_reference' => array(
						'title' => __( 'Widget Reference', 'woo_partpay' ),
						'type' => 'text',
						'label' => __( 'Widget Reference', 'woo_partpay' ),
						'default' => __( '', 'woo_partpay' )
					),
					'enable_order_notes' => array(
						'title' => __( 'Order Page Notes', 'woo_partpay' ),
						'type' => 'checkbox',
						'label' => __( 'Enable Order Detail Page Notes', 'woo_partpay' ),
						'default' => 'no'
					)
				);
			} // End init_form_fields()

			/**
			 * Init JS Scripts Options
			 *
			 * @since 1.2.1
			 */
			public function init_scripts_js() {
				//use WP native jQuery
				wp_enqueue_script("jquery");

			}

			/**
			 * Init Scripts Options
			 *
			 * @since 1.2.1
			 */
			public function init_scripts_css() {
			}

			/**
			 * Init Environment Options
			 *
			 * @since 1.2.3
			 */
			public function init_environment_config() {
				if ( empty( $this->environments ) ) {
					//config separated for ease of editing
					require( 'config/config.php' );
					$this->environments = $environments;
				}
			}

			/**
			 * Admin Panel Options
			 *
			 * @since 1.0.0
			 */
			public function admin_options() {
				?>
				<h3><?php _e('Payflex Gateway', 'woo_partpay'); ?></h3>

				<table class="form-table">
					<?php
					// Generate the HTML For the settings form.
					$this->generate_settings_html();
					?>
				</table><!--/.form-table-->
				<?php
			} // End admin_options()


/**
			 * Display payment options on the checkout page
			 *
			 * @since 1.0.0
			 */
			public function payment_fields() {

				global $woocommerce;
				$settings = get_option('woocommerce_payflex_settings');
				if(isset($settings['enable_checkout_widget']) && $settings['enable_checkout_widget'] == 'yes'){
						echo '<style>.elementor{max-width:100% !important}';
						echo 'html {
							-webkit-font-smoothing: antialiased!important;
							-moz-osx-font-smoothing: grayscale!important;
							-ms-font-smoothing: antialiased!important;
						}
						.md-stepper-horizontal {
							display:table;
							width:100%;
							margin:0 auto;
							background-color:transparent;
						}
						.md-stepper-horizontal .md-step {
							display:table-cell;
							position:relative;
							padding:0;
						}
						.md-stepper-horizontal .md-step .md-step-circle {
							width:30px;
							height:30px;
							margin:0 auto;
							border-radius: 50%;
							text-align: center;
							line-height:30px;
							font-size: 16px;
							font-weight: 600;
							color:#FFFFFF;
						}
						.md-stepper-horizontal .md-step .md-step-title {
							margin-top:16px;
							font-size:16px;
							font-weight:600;
						}
						.md-stepper-horizontal .md-step .md-step-title,
						.md-stepper-horizontal .md-step .md-step-optional {
							text-align: center;
							color : #002751;
						}
						.md-stepper-horizontal .md-step .md-step-optional {
							font-size:12px;
						}
						.payflex_description{
							font-size:16px;text-align:center;
							margin-top:17px;
						}
						.fontcolor{
							color : #002751;
						}
						</style>';
						$ordertotal = $woocommerce->cart->total;
						$installment =  round(($ordertotal / 4),2);
						echo '<div class="fontcolor" style="font-size:16px;text-align:center">Four interest-free payments totalling R'.$ordertotal.'</div>';
						echo '<div class="md-stepper-horizontal orange">
							<div class="md-step active">
							<div class="md-step-title">R'.$installment.'</div>
							<div class="md-step-circle"><span><img src ="'.WP_PLUGIN_URL . "/" . plugin_basename( dirname(__FILE__)) . '/PIE-CHART-01.png'.'"></span></div>
							<div class="md-step-optional">1st instalment</div>
							</div>
							<div class="md-step active">
							<div class="md-step-title">R'.$installment.'</div>
							<div class="md-step-circle"><span><img src ="'.WP_PLUGIN_URL . "/" . plugin_basename( dirname(__FILE__)) . '/PIE-CHART-02.png'.'"></span></div>
							<div class="md-step-optional">2 weeks later</div>
							</div>
							<div class="md-step active">
							<div class="md-step-title">R'.$installment.'</div>
							<div class="md-step-circle"><span><img src ="'.WP_PLUGIN_URL . "/" . plugin_basename( dirname(__FILE__)) . '/PIE-CHART-03.png'.'"></span></div>
							<div class="md-step-optional">4 weeks later</div>
							</div>
							<div class="md-step active">
							<div class="md-step-title">R'.$installment.'</div>
							<div class="md-step-circle"><span><img src ="'.WP_PLUGIN_URL . "/" . plugin_basename( dirname(__FILE__)) . '/PIE-CHART-04.png'.'"></span></div>
							<div class="md-step-optional">6 weeks later</div>
							</div>
						</div>
					<div class="payflex_description fontcolor">You will be redirected to Payflex when you click on place order.</div>
					';
				}else{
					if ($this->settings['testmode'] != 'production') : ?><?php _e('TEST MODE ENABLED', 'woo_partpay'); ?><?php endif;
					$arr = array( 'br' => array(), 'p' => array());
					if ($this->description) { echo wp_kses('<p>'.$this->description.'</p>',$arr); }
				}
			
			}
			/**
			 * Request an order token from Partpay
			 *
			 * @return  string or boolean false if no token generated
			 * @since 1.0.0
			 */
			public function get_partpay_authorization_code() {

				$access_token = get_transient( 'partpay_access_token');
				$this->log('Access token from cache is ' . $access_token);
				if( false !== $access_token && !empty($access_token)){
					$this->log('returning token ' . $access_token);
					return $access_token;
				}
				if( false === $this->apiKeysAvailable() ){
					$this->log('no api keys available');
					return false;
				}
				$this->log('Getting new token');
				$AuthURL = $this->environments[ $this->settings['testmode'] ]['auth_url'];
				$AuthBody = [
					'client_id'  => $this->settings['client_id'],
					'client_secret' => $this->settings['client_secret'],
					'audience'=>$this->environments[ $this->settings['testmode'] ]['auth_audience'],
					'grant_type'=>'client_credentials'
				];
				$AuthBody = wp_json_encode( $AuthBody );
				$headers = array( 'Content-Type' => 'application/json');
				$AuthBody = json_decode(json_encode($AuthBody), true);     
				
				$response = wp_remote_post($AuthURL,array('body' => $AuthBody,'headers'=> $headers ));
				$body = json_decode(wp_remote_retrieve_body($response));
				if ( $response['response']['code'] != '401' ) {
					//store this in cache
					$this->log('Storing new token in cache ' . $body->access_token . ' which is valid for ' . $body->expires_in . ' seconds');
					set_transient('partpay_access_token', $body->access_token,((int)$body->expires_in - 120 ));
					return $body->access_token;
				}else {
					return false;
				}
			}

			private function apiKeysAvailable(){

				if( empty($this->settings['client_id']) || empty($this->settings['client_secret']) ){
					$this->log('API keys not available.');
					return false;
				}

				return true;
			}


			public function update_payment_limits(){
				// Get existing limits
				$settings = get_option('woocommerce_payflex_settings');

				if( false === $this->apiKeysAvailable() ){
					return false;
				}

				$this->log( 'Updating payment limits requested');
				if(!empty($this->configurationUrl)){	
					$response = wp_remote_get($this->configurationUrl,array('headers'=>array('Authorization' => 'Bearer ' . $this->get_partpay_authorization_code())));
					$body = json_decode(wp_remote_retrieve_body($response));

					$this->log( 'Updating payment limits response: '.print_r($body,true) );

					if ($response['response']['code'] == 200) {
						$settings['partpay-amount-minimum'] = $body->minimumAmount;
						$settings['partpay-amount-maximum'] = $body->maximumAmount;
					}

					update_option('woocommerce_payflex_settings',$settings);
				}
				$this->init_settings();
			

		}


			/**
			 * Process the payment and return the result
			 * - redirects the customer to the pay page
			 *
			 * @param int $order_id
			 *
			 * @since 1.0.0
			 * @return array
			 */
			public function process_payment( $order_id ) {

				if( function_exists("wc_get_order") ) {
					$order = wc_get_order( $order_id );
				}
				else {
					$order = new WC_Order( $order_id );
				}

				// Get the authorization token
				$access_token = $this->get_partpay_authorization_code();

				//Process here
				$orderitems = $order->get_items();
				$items = array();
				$i = 0;
				if (count($orderitems)) {
					foreach ($orderitems as $item) {

						$i++;
						// get SKU
						if ($item['variation_id']) {

							if(function_exists("wc_get_product")) {
								$product = wc_get_product($item['variation_id']);
							}else {
								$product = new WC_Product($item['variation_id']);
							}
						} else {

							if(function_exists("wc_get_product")) {
								$product = wc_get_product($item['product_id']);
							} else {
								$product = new WC_Product($item['product_id']);
							}
						}



						if ($i ==  count($orderitems) ) {
							$product =
								$items[] = array(


								'{
									"name":"'. esc_html( $item['name'] ) . $i.'",
									"sku":"'.$product->get_sku().'",
									"quantity":"'. $item['qty'].'",
									"price":"'. number_format(($item['line_subtotal'] / $item['qty']),2,'.','') .'"
								}'

							);
						}else{
							$product =
							$items[] = array(


								'{
									"name":"'. esc_html( $item['name'] ) . $i.'",
									"sku":"'.$product->get_sku().'",
									"quantity":"'. $item['qty'].'",
									"price":"'. number_format(($item['line_subtotal'] / $item['qty']),2,'.','') .'"
								},'

							);
						}
					}
				}

				//calculate total shipping amount
				if( method_exists($order, 'get_shipping_total') ){
					//WC 3.0
					$shipping_total = $order->get_shipping_total();
				} else {
					//WC 2.6.x
					$shipping_total = $order->get_total_shipping();
				}

				$OrderBodyString = '{
				"amount": '.number_format( $order->get_total(), 2, '.', '').',
				"consumer": {
				"phoneNumber":  "'.$order->billing_phone.'",
					"givenNames":  "'.$order->billing_first_name.'",
					"surname":  "'.$order->billing_last_name.'",
					"email":  "'.$order->billing_email.'"
				},
				"billing": {
					"addressLine1":"'.$order->billing_address_1.'",
					"addressLine2": "'.$order->billing_address_2.'",
					"suburb": "'.$order->billing_city.'",
					"postcode": "'.$order->billing_postcode.'"
				},
				"shipping": {
					"addressLine1": "'.$order->shipping_address_1.'",
					"addressLine2": " '.$order->shipping_address_2.'",
					"suburb":  "'.$order->shipping_city.'",
					"postcode": "'.$order->shipping_postcode.'"
				},
				"description": "string",
				"items": [';

				foreach ($items as $item) {
					$OrderBodyString .=  $item[0] ;
				}

				$OrderBodyString .= '],
				"merchant": {
					"redirectConfirmUrl": "'. $this->get_return_url( $order ).'&status=confirmed",
					"redirectCancelUrl": "'. $this->get_return_url( $order ).'&status=cancelled"
				},
				"merchantReference": "'.$order_id.'",
				"taxAmount": '.$order->get_total_tax().',
				"shippingAmount":'.$shipping_total.'
				}';

				$OrderBody =  json_decode($OrderBodyString);

				$APIURL = $this->orderurl.'/productSelect';

				$order_args = array(
					'method' => 'POST',
					'headers' => array( 'Content-Type' => 'application/json', 'Authorization' => 'Bearer '. $access_token ),
					'body' =>  json_encode( $OrderBody),
					'timeout' => 30
				);

				$this->log( 'POST Order request: '.print_r($order_args,true) );

				$order_response = wp_remote_post($APIURL, $order_args);

				$order_body = json_decode(wp_remote_retrieve_body($order_response));

				if ($access_token == false) {
					// Couldn't generate token
					$order->add_order_note(__('Unable to generate the order token. Payment couldn\'t proceed.', 'woo_partpay'));
					wc_add_notice(__('Sorry, there was a problem preparing your payment.', 'woo_partpay'),'error');
					return array(
						'result' => 'failure',
						'redirect' => $order->get_checkout_payment_url(true)
					);

				} else {
					$this->log( 'Created Payflex OrderId: '.print_r($order_body->orderId,true) );
					// Order token successful, save it so we can confirm it later
					update_post_meta($order_id,'_partpay_order_token',$order_body->token);
					update_post_meta($order_id,'_partpay_order_id',$order_body->orderId);
					update_post_meta($order_id,'_order_redirectURL', $order_body->redirectUrl);
					$savedId = get_post_meta($order_id, '_partpay_order_id', true);
					$this->log( 'Saved '.$savedId.' into post meta' );
				}

				$redirect = $order->get_checkout_payment_url(true);

				return array(
					'result' 	=> 'success',
					'redirect'	=> $redirect
				);

			}

			/**
			 * Update status after API redirect back to merchant site
			 *
			 * @since 1.0.0
			 */
			public function receipt_page($order_id) {

				if( function_exists("wc_get_order") ) {
					$order = wc_get_order( $order_id );
				}
				else {
					$order = new WC_Order( $order_id );
				}


				$redirectURL = get_post_meta($order_id,'_order_redirectURL');

				//Update order status if it isn't already
				$is_pending = false;
				if ( function_exists("has_status") ) {
					$is_pending = $order->has_status('pending');
				}
				else {
					if( $order->get_status() == 'pending' ) {
						$is_pending = true;
					}
				}

				if ( !$is_pending ) {
					$order->update_status('pending');
				}

				//Redirect to Partpay checkout
				header('Location: '.$redirectURL[0]);
			}

			public function get_partpay_order_id($order_id){

				$partpay_order_id = get_post_meta($order_id, '_partpay_order_id', true);

				if( empty($partpay_order_id) )
					return false;
			}

			/**
			 * @param $order_id
			 */
			public function payment_callback($order_id) {

				if( function_exists("wc_get_order") ) {
					$order = wc_get_order( $order_id );
				}
				else {
					$order = new WC_Order( $order_id );
				}

				// //save the order id

				$order_token = get_post_meta($order_id, '_partpay_order_token', true);
				$partpay_order_id =  get_post_meta($order_id, '_partpay_order_id', true);

				$this->log( sprintf('Attempting to set order status for %s, payflex orderId: %s, token: %s', $order_id, $partpay_order_id, $order_token) );

				if ( !empty($partpay_order_id) ) {

					//Update status for order
					if (isset($_GET['status'])) {

						$query_string = sanitize_text_field($_GET['status']);

						if ($query_string != NULL) {
							if ($query_string == 'confirmed') {

								$order->add_order_note(sprintf(__('Payment approved. Payflex Order ID: '.$partpay_order_id.' ','woo_partpay') ));
								$order->payment_complete($partpay_order_id);
								wc_empty_cart();

							}elseif ($query_string == 'cancelled') {

								$order->add_order_note(sprintf(__('Payflex payment is pending approval. Payflex Order ID: '.$partpay_order_id.' ','woo_partpay') ));
								$order->update_status( 'cancelled' );


							} elseif ($query_string == 'failure') {

								$order->add_order_note(sprintf(__('Payflex payment declined. Order ID from Payflex: '.$partpay_order_id.' ','woo_partpay') ));
								$order->update_status('failed');
							}

						}else{
							$order->update_status('pending');
						}
					}

				}

			}

			/**
			 * Check whether the cart amount is within payment limits
			 *
			 * @param  array $gateways Enabled gateways
			 * @return  array Enabled gateways, possibly with PartPay removed
			 * @since 1.0.0
			 */
			public function check_cart_within_limits($gateways) {

				global $woocommerce;
				$total = isset($woocommerce->cart->total) ? $woocommerce->cart->total : 0;

				$access_token = $this->get_partpay_authorization_code();
				$config_response_transistent = get_transient( 'payflex_configuration_response');
				$api_url = $this->configurationUrl;
				if( false !== $config_response_transistent && !empty($config_response_transistent)){
					$order_response  = $config_response_transistent;
					$order_body = json_decode($order_response);
				}else{

					$order_args = array(
						'method' => 'GET',
						'headers' => array( 'Content-Type' => 'application/json', 'Authorization' => 'Bearer '. $access_token ),
						'timeout' => 30
					);
					$order_response = wp_remote_post($api_url, $order_args);
					$order_response = wp_remote_retrieve_body($order_response);
					$order_body = json_decode($order_response);
					set_transient('payflex_configuration_response', $order_response,86400);
				}
				if($order_response){
				
					$pbi = ($total >= $order_body->minimumAmount && $total <= $order_body->maximumAmount);

					if (!$pbi) {
						unset($gateways['partpay']);
					}

				}	
				return $gateways;				

			}


			/**
			 * Can the order be refunded?
			 *
			 * @param  WC_Order $order
			 * @return bool
			 */
			public function can_refund_order( $order ) {
				return $order && $order->get_transaction_id();
			}

			/**
			 * Process a refund if supported
			 *
			 * @param  WC_Order $order
			 * @param  float $amount
			 * @param  string $reason
			 * @return  boolean True or false based on success
			 */
			public function process_refund( $order_id, $amount = null, $reason = '' ) {

				$this->log(sprintf('Attempting refund for order_id: %s', $order_id));

				$partpay_order_id = get_post_meta($order_id, '_partpay_order_id', true);

				$this->log(sprintf('Attempting refund for Payflex OrderId: %s', $partpay_order_id));

				$order = new WC_Order($order_id);

				if( empty($partpay_order_id) ){
					$order->add_order_note(sprintf(__('There was an error submitting the refund to Payflex.', 'woo_partpay')));
					return false;
				}

				$access_token = $this->get_partpay_authorization_code();
				$random_string = wp_generate_password(8, false, false);
				error_log('partpay orderId2'.$partpay_order_id);
				$refund_args = array(
					'headers' => array(
						'Content-Type' => 'application/json',
						'Authorization' => 'Bearer '. $access_token
					),
					'body'  =>  json_encode(array(
						'requestId' => 'Order #'.$order_id .'-'. $random_string,
						'amount' => $amount,
						'merchantRefundReference' => 'Order #'.$order_id .'-'. $random_string
					))
				);
				error_log('partpay orderId3'.$partpay_order_id);
				$refundOrderUrl = $this->orderurl.'/'.$partpay_order_id.'/refund';

				$refund_response = wp_remote_post($refundOrderUrl, $refund_args);
				$refund_body = json_decode(wp_remote_retrieve_body($refund_response));

				$this->log( 'Refund body: '.print_r($refund_body,true) );
				error_log('partpay orderId3'.$partpay_order_id);
				$responsecode = intval( $refund_response['response']['code'] );

				if ($responsecode == 201 || $responsecode == 200) {
					$order->add_order_note(sprintf(__('Refund of $%s successfully sent to PayFlex.', 'woo_partpay'),$amount));
					return true;
				} else {
					if ($responsecode == 404) {
						$order->add_order_note(sprintf(__('Order not found on Payflex.', 'woo_partpay')));
					}else{
						$order->add_order_note(sprintf(__('There was an error submitting the refund to Payflex.', 'woo_partpay')));
					}
					return false;
				}


			}


			/**
			 * Logging method
			 * @param  string $message
			 */
			public static function log( $message ) {

				if ( self::$log_enabled ) {
					if ( empty( self::$log ) ) {
						self::$log = new WC_Logger();
					}
					self::$log->add( 'partpay', $message );
				}
			}

			/**
			 * @param $order_id
			 */
			public function create_refund( $order_id ) {

				$order = new WC_Order( $order_id );
				$order_refunds = $order->get_refunds();


				if( !empty($order_refunds) ) {
					$refund_amount = $order_refunds[0]->get_refund_amount();
					if( !empty($refund_amount) ) {
						$this->process_refund($order_id, $refund_amount, "Admin Performed Refund");
					}
				}
			}

			/**
			 * Check the order status of all orders that didn't return to the thank you page or marked as Pending by PartPay
			 *
			 * @since 1.0.0
			 */
			public function check_pending_abandoned_orders() {

				// $pending_orders = get_posts(array('post_type'=>'shop_order','post_status'=>'wc-pending',
					$pending_orders = get_posts(array('post_type'=>'shop_order',
					'post_status'=> array('wc-pending','wc-failed','wc-cancelled'),
					'date_query' => array(
						array(
							'after' => '24 hours ago'
							)
						)
					));
				$this->log( 'checking '. count($pending_orders) . " orders" );
				foreach ( $pending_orders as $pending_order ) {
					if(function_exists("wc_get_order") ) {
						$order = wc_get_order( $pending_order->ID );
					}
					else {
						$order = new WC_Order( $pending_order->ID );
					}

					//skip all orders that are not PartPay
					$payment_method = get_post_meta( $pending_order->ID, '_payment_method', true );
					if( $payment_method != "payflex" ) {
						$this->log( 'ignoring order not from payflex: '. $payment_method );
						continue;
					}

					$partpay_order_id = get_post_meta($pending_order->ID,'_partpay_order_id',true);

					// Check if there's a stored order token. If not, it's not an PartPay order.
					if (!$partpay_order_id)  {
						$this->log( 'No Payflex OrderId for Order '.$pending_order->ID );
						continue;
					}

					$this->log( 'Checking abandoned order for WC Order ID '.$order->ID.', Payflex ID '.$partpay_order_id);

					$response = wp_remote_get(
						$this->orderurl.'/'.$partpay_order_id,
						array(
							'headers'=>array(
								'Authorization' => 'Bearer ' . $this->get_partpay_authorization_code(),
							)
						)
					);
					$body = json_decode(wp_remote_retrieve_body($response));

					$this->log( 'Checking abandoned order result: '.print_r($body,true) );

					$response_code = wp_remote_retrieve_response_code($response);
					$settings = get_option('woocommerce_payflex_settings');
					// Check status of order
					if ($response_code == 200) {
						// Check status of order
						if ($body->orderStatus == "Approved") {
							$order->add_order_note(sprintf(__('Checked payment status with Payflex. Payment approved. Payflex Order ID: %s','woo_partpay'),$partpay_order_id));
							$order->payment_complete($pending_order->orderId);
						} elseif ($body->orderStatus == "Created") {
							if($settings['enable_order_notes'] == 'yes'){
								$order->add_order_note(__('Checked payment status with Payflex. Still pending approval.','woo_partpay'));
							}
							$order->update_status( 'pending' );
						} elseif ($body->orderStatus == 'Abandoned' ){
							$order->add_order_note(sprintf(__('Checked payment status with Payflex. Payment %s. Payflex Order ID: %s','woo_partpay'),strtolower($body->orderStatus),$partpay_order_id));
							$order->update_status( 'cancelled' );
						} elseif ($body->orderStatus == 'Declined' ){
							$order->add_order_note(sprintf(__('Checked payment status with Payflex. Payment %s. Payflex Order ID: %s','woo_partpay'),strtolower($body->orderStatus),$partpay_order_id));
							$order->update_status( 'cancelled' );
						} else {
							$order->add_order_note(sprintf(__('Checked payment status with Payflex. Payment %s. Payflex Order ID: %s','woo_partpay'),strtolower($body->orderStatus),$partpay_order_id));
							$order->update_status( 'failed' );
						}
					} else {
						$order->add_order_note(sprintf(__('Tried to check payment status with Payflex. Unable to access API. Repsonse code is %s Payflex Order ID: %s','woo_partpay'),$response_code,$partpay_order_id));
					}

				}

			}

		}


		/**
		 * Add the Partpay gateway to WooCommerce
		 *
		 * @param  array $methods Array of Payment Gateways
		 * @return  array Array of Payment Gateways
		 * @since 1.0.0
		 **/
		function add_partpay_gateway( $methods ) {
			$methods[] = 'WC_Gateway_PartPay';
			return $methods;
		}
		add_filter('woocommerce_payment_gateways', 'add_partpay_gateway' );


		/**
		 * Check for the CANCELLED payment status
		 * We have to do this before the gateway initialises because WC clears the cart before initialising the gateway
		 *
		 * @since 1.0.0
		 */
		function partpay_check_for_cancelled_payment() {
			// Check if the payment was cancelled
			if (isset($_GET['status']) && $_GET['status'] == "cancelled" && isset($_GET['key']) && isset($_GET['token'])) {

				$gateway = WC_Gateway_PartPay::instance();
				$key = sanitize_text_field( $_GET['key'] );
				$order_id = wc_get_order_id_by_order_key($key);


				if( function_exists("wc_get_order") ) {
					$order = wc_get_order( $order_id );
				}
				else {
					$order = new WC_Order( $order_id );
				}

				if ($order) {

					$partpay_order_id = get_post_meta($order_id, '_partpay_order_id', true);
					$obj = new WC_Gateway_PartPay();
					$response = wp_remote_get(
						$obj->orderurl.'/'.	$partpay_order_id,
						array(
							'headers'=>array(
								'Authorization' => 'Bearer ' . $obj->get_partpay_authorization_code(),
							)
						)
					);
					$body = json_decode(wp_remote_retrieve_body($response));
					if ($body->orderStatus != "Approved") {
						$gateway->log( 'Order '.$order_id.' payment cancelled by the customer while on the Payflex checkout pages.' );
						$order->add_order_note(__('Payment cancelled by the customer while on the Payflex checkout pages.', 'woo_partpay'));

						if( method_exists($order, "get_cancel_order_url_raw") ) {
							wp_redirect($order->get_cancel_order_url_raw());
						}
						else {

							wp_redirect($order->get_cancel_order_url());
						}
						exit;
					}
					$redirect = $order->get_checkout_payment_url(true);

					return array(
						'result' 	=> 'success',
						'redirect'	=> $redirect
					);
				}
			}
		}
		add_action('template_redirect','partpay_check_for_cancelled_payment');

		function partpay_info_product_page_widget() {
			global $post;
			if( function_exists("wc_get_product") ) {
				$product = wc_get_product($post->ID);
			}
			else {
				$product = new WC_Product($post->ID);
			}
			$price = ($product) ? $product->get_price() : 0;
			$settings = get_option('woocommerce_payflex_settings');

			if (!isset($settings['enable_product_widget']) || $settings['enable_product_widget'] !== 'yes' || $settings['is_using_page_builder'] == 'yes')
				return;

			$merchantSlug = '/';
			if (isset($settings['merchant_widget_reference']))
				$merchantSlug .= $settings['merchant_widget_reference'] . '/';

			$return_html = '<div><script async src="https://widgets.payflex.co.za';
			$return_html .= $merchantSlug;
			$return_html .= 'partpay-widget-0.1.1.js?type=calculator&amount=';
			$return_html .= $price;
			$return_html .= '&min='. $settings['partpay-amount-minimum'];
			$return_html .= '&max='. $settings['partpay-amount-maximum'];
			$return_html .= '" type="application/javascript"></script></div>';
			$allowed_html = array(
				'script'      => array(
					'src'  => array(),
					'async' => array(),
				),
				'div'     => array()
			);
			echo wp_kses( $return_html, $allowed_html );
		}

		add_action('woocommerce_single_product_summary','partpay_info_product_page_widget',15);
		
		function partpay_info_product_page_widget_shortcode() {
			global $post;
			if( function_exists("wc_get_product") ) {
				$product = wc_get_product($post->ID);
			}
			else {
				$product = new WC_Product($post->ID);
			}
			$price = ($product) ? $product->get_price() : 0;
			$settings = get_option('woocommerce_payflex_settings');

			if (!isset($settings['enable_product_widget']) || $settings['enable_product_widget'] !== 'yes' || $settings['is_using_page_builder'] !== 'yes')
				return;

			$merchantSlug = '/';
			if (isset($settings['merchant_widget_reference']))
				$merchantSlug .= $settings['merchant_widget_reference'] . '/';

			$return_html = '<div><script async src="https://widgets.payflex.co.za';
			$return_html .= $merchantSlug;
			$return_html .= 'partpay-widget-0.1.1.js?type=calculator&amount=';
			$return_html .= $price;
			$return_html .= '&min='. $settings['partpay-amount-minimum'];
			$return_html .= '&max='. $settings['partpay-amount-maximum'];
			$return_html .= '" type="application/javascript"></script></div>';
			$allowed_html = array(
				'script'      => array(
					'src'  => array(),
					'async' => array(),
				),
				'div'     => array()
			);
			return wp_kses( $return_html, $allowed_html );
		}

		add_shortcode('payflex_widget','partpay_info_product_page_widget_shortcode');
			
		function partpay_variation_price_widget( $price, $variation ) {
			$return_html = $price;
			$settings = get_option('woocommerce_payflex_settings');

			if (!isset($settings['enable_product_widget']) || $settings['enable_product_widget'] !== 'yes')
				return;
			$merchantSlug = '/';

			if (!isset($settings['merchant_widget_reference']))
				$merchantSlug .= $settings['merchant_widget_reference'] . '/';

			$amount = ($variation) ? $variation->get_price() : 0;

			$return_html .= '<div><script async src="https://widgets.payflex.co.za';
			$return_html .= $merchantSlug;
			$return_html .= 'partpay-widget-0.1.1.js?type=calculator&amount=';
			$return_html .= $amount;
			$return_html .= '" type="application/javascript"></script></div>';
			$allowed_html = array(
				'script'      => array(
					'src'  => array(),
					'async' => array(),
				),
				'div'     => array()
			);
			return wp_kses( $return_html, $allowed_html );
		}

		add_filter( 'woocommerce_variation_price_html', 'partpay_variation_price_widget', 10, 2);
		add_filter( 'woocommerce_variation_sale_price_html', 'partpay_variation_price_widget', 10, 2);



		/**
		 * Call the cron task related methods in the gateway
		 *
		 * @since 1.0.0
		 **/
		function partpay_do_cron_jobs() {
			$gateway = WC_Gateway_Partpay::instance();
			$gateway->check_pending_abandoned_orders();
			$gateway->update_payment_limits();
		}
		add_action('partpay_do_cron_jobs','partpay_do_cron_jobs');
		add_action('init', function() {
			if (!wp_next_scheduled ( 'partpay_do_cron_jobs' )) {
				wp_schedule_event(time(),'twominutes','partpay_do_cron_jobs');
			}
		});
	}
		

	/* WP-Cron activation and schedule setup */

	/**
	 * Schedule PartPay WP-Cron job
	 *
	 * @since 1.0.0
	 **/
	function partpay_create_wpcronjob() {
		$timestamp = wp_next_scheduled('partpay_do_cron_jobs');
		if ($timestamp == false) {
			wp_schedule_event(time(),'twominutes','partpay_do_cron_jobs');
		}
	}
	register_activation_hook( __FILE__, 'partpay_create_wpcronjob' );

	/**
	 * Delete PartPay WP-Cron job
	 *
	 * @since 1.0.0
	 **/
	function partpay_delete_wpcronjob(){
		wp_clear_scheduled_hook( 'partpay_do_cron_jobs' );
	}
	register_deactivation_hook( __FILE__, 'partpay_delete_wpcronjob' );

	/**
	 * Add a new WP-Cron job scheduling interval of every 2 minutes
	 *
	 * @param  array $schedules
	 * @return array Array of schedules with 2 minutes added
	 * @since 1.0.0
	 **/
	function partpay_add_two_minute_schedule( $schedules ) {
		$schedules['twominutes'] = array(
			'interval' => 120,  // seconds
			'display' => __( 'Every 2 minutes', 'woo_partpay' )
		);
		return $schedules;
	}
	add_filter('cron_schedules', 'partpay_add_two_minute_schedule');

}